import itertools
import numpy as np


def confidence(F, alpha=.05):
    r"""
    Constructs a Dvoretzky-Kiefer-Wolfowitz confidence band for the eCDF.

    Parameters
    ----------
    F : array_like
        The empirical distributions
    alpha : float
        Set alpha for a (1 - alpha) % confidence band.

    Notes
    -----
    Based on the DKW inequality.

    .. math:: P \left( \sup_x \left| F(x) - \hat(F)_n(X) \right| > \epsilon \right) \leq 2e^{-2n\epsilon^2}

    References
    ----------
    Wasserman, L. 2006. `All of Nonparametric Statistics`. Springer.

    """
    eps = np.sqrt(np.log(2./alpha) / (2 * len(F)))
    return np.clip(F - eps, 0, 1), np.clip(F + eps, 0, 1)


def core_one(n, r):
    return np.kron(np.ones([1, n, 1]), np.eye(r)[:, None, :])


def get_cdf(x):
    _x = np.array(x, copy=True)
    _x.sort()
    _y = np.linspace(1./len(_x), 1, len(_x))

    _x = np.r_[-np.inf, _x]
    _y = np.r_[0, _y]

    def cdf(z):
        return _y[np.searchsorted(_x, z, 'right') - 1]

    return cdf


def get_partial(Y, n):
    Q = Y[0][0, n[0], :]
    for i in range(1, len(Y)):
        Q = np.einsum('q,qp->p', Q, Y[i][:, n[i], :])
    return Q


def ind2poi(I, a, b, n):
    """Transforms multiindices (samples) into points of the uniform grid."""
    if isinstance(a, list): a = np.array(a)
    if isinstance(b, list): b = np.array(b)
    if isinstance(n, list): n = np.array(n)

    if len(I.shape) == 1:
        # If we have only one multiindex
        t = I * 1. / (n - 1)
        x = t * (b - a) + a
        return x

    A = np.repeat(a.reshape((1, -1)), I.shape[0], axis=0)
    B = np.repeat(b.reshape((1, -1)), I.shape[0], axis=0)
    N = np.repeat(n.reshape((1, -1)), I.shape[0], axis=0)
    T = I * 1. / (N - 1)
    X = T * (B - A) + A
    return X


def ind2str(i):
    """Transforms array of int like [1, 2, 3] into string like '1-2-3'."""
    return '-'.join([str(int(v)) for v in i])


def kron(a, b):
    return np.kron(a, b)


def lhs(shape, samples):
    d = len(shape)
    indices = np.empty((samples, d), dtype=int)
    for i, sh in enumerate(shape):
        part1 = np.repeat(np.arange(sh), samples // sh)
        part2 = np.random.choice(sh, samples-len(part1), replace=False)
        indices[:, i] = np.concatenate([part1, part2])
        np.random.shuffle(indices[:, i])
    return indices


def orthogonalize(Z, k):
    # Z = [G.copy() for G in Y]
    L = np.array([[1.]])
    R = np.array([[1.]])
    for i in range(0, k):
        G = reshape(Z[i], [-1, Z[i].shape[2]])
        Q, R = np.linalg.qr(G, mode='reduced')
        Z[i] = reshape(Q, Z[i].shape[:-1] + (Q.shape[1], ))
        G = reshape(Z[i+1], [Z[i+1].shape[0], -1])
        Z[i+1] = reshape(np.dot(R, G), (R.shape[0], ) + Z[i+1].shape[1:])
    for i in range(len(Z)-1, k, -1):
        G = reshape(Z[i], [Z[i].shape[0], -1])
        L, Q = scipy.linalg.rq(G, mode='economic', check_finite=False)
        Z[i] = reshape(Q, (Q.shape[0], ) + Z[i].shape[1:])
        G = reshape(Z[i-1], [-1, Z[i-1].shape[2]])
        Z[i-1] = reshape(np.dot(G, L), Z[i-1].shape[:-1] + (L.shape[1], ))
    return Z


def reshape(a, sz):
    return np.reshape(a, sz, order='F')


def second_order_2_TT(A, i, j, shapes):
    if i > j: # Так не должно быть
        j, i = i, j
        A = A.T

    U, V = skeleton(A)
    r = U.shape[1]

    core1 = U.reshape(1, U.shape[0], r)
    core2 = V.reshape(r, V.shape[1], 1)

    cores = []
    for num, n in enumerate(shapes):
        if num < i or num > j:
            cores.append(np.ones([1, n, 1]))
        if num == i:
            cores.append(core1)
        if num == j:
            cores.append(core2)
        if i < num < j:
            cores.append(core_one(n, r))

    return cores


def skeleton(a, eps=1.E-10, r=int(1e12), hermitian=False):
    u, s, v = np.linalg.svd(a, full_matrices=False,
        compute_uv=True, hermitian=hermitian)
    r = min(r, sum(s>eps))
    un = u[:, :r]
    sn = np.diag(np.sqrt(s[:r]))
    vn = v[:r]
    return un @ sn, sn @ vn


def sum_many(tensors, e=1.E-10, rmax=None, freq_trunc=15):
    cores = tensors[0]
    for i, t in enumerate(tensors[1:]):
        cores = teneva.add(cores, t)
        if (i+1) % freq_trunc == 0:
            cores = teneva.truncate(cores, e=e)
    return teneva.truncate(cores, e=e, rmax=rmax)


def svd_truncated(M, delta, rmax=None):
    if M.shape[0] <= M.shape[1]:
        cov = M.dot(M.T)
        singular_vectors = 'left'
    else:
        cov = M.T.dot(M)
        singular_vectors = 'right'

    if np.linalg.norm(cov) < 1e-14:
        return np.zeros([M.shape[0], 1]), np.zeros([1, M.shape[1]])

    w, v = np.linalg.eigh(cov)
    w[w < 0] = 0
    w = np.sqrt(w)
    svd = [v, w]
    idx = np.argsort(svd[1])[::-1]
    svd[0] = svd[0][:, idx]
    svd[1] = svd[1][idx]
    S = svd[1]**2
    where = np.where(np.cumsum(S[::-1]) <= delta**2)[0]
    if len(where) == 0:
        rank = max(1, int(np.min([rmax, len(S)])))
    else:
        rank = max(1, int(np.min([rmax, len(S) - 1 - where[-1]])))
    left = svd[0]
    left = left[:, :rank]

    if singular_vectors == 'left':
        M2 = ((1. / svd[1][:rank])[:, np.newaxis]*left.T).dot(M)
        left = left*svd[1][:rank]
    else:
        M2 = M.dot(left)
        left, M2 = M2, left.T

    return left, M2


def str2ind(s):
    """Transforms string like '1-2-3' into array of int like [1, 2, 3]."""
    return np.array([int(v) for v in s.split('-')], dtype=int)


def tt_sample(shape, k):
    def one_mode(sh1, sh2, rng):
        res = []
        if len(sh2) == 0:
            lhs_1 = lhs(sh1, k)
            for n in range(rng):
                for i in lhs_1:
                    res.append(np.concatenate([i, [n]]))
            len_1, len_2 = len(lhs_1), 1
        elif len(sh1) == 0:
            lhs_2 = lhs(sh2, k)
            for n in range(rng):
                for j in lhs_2:
                    res.append(np.concatenate([[n], j]))
            len_1, len_2 = 1, len(lhs_2)
        else:
            lhs_1 = lhs(sh1, k)
            lhs_2 = lhs(sh2, k)
            for n in range(rng):
                for i, j in itertools.product(lhs_1,  lhs_2):
                    res.append(np.concatenate([i, [n], j]))
            len_1, len_2 = len(lhs_1), len(lhs_2)
        return res, len_1, len_2

    idx = [0]
    idx_many = []
    pnts_many = []

    for i in range(len(shape)):
        pnts, len_1, len_2 = one_mode(shape[:i], shape[i+1:], shape[i])
        pnts_many.append(pnts)
        idx.append(idx[-1] + len(pnts))
        idx_many.append(len_2)

    return np.vstack(pnts_many), np.array(idx), np.array(idx_many)


def tt_svd(t, eps=1E-10, max_r=int(1e12)):
    A = t = np.asanyarray(t)
    r = 1
    res = []
    for sh in t.shape[:-1]:
        A = A.reshape(r*sh, -1)
        G, A = skeleton(A, eps=eps, r=max_r)
        G = G.reshape(r, sh, -1)
        res.append(G)
        r = G.shape[-1]
    res.append(A.reshape(r, t.shape[-1], 1))
    return res


def tt_svd_incomplete(I, Y, idx, idx_many, rank, eps_skel=1e-10):
    shapes = np.max(I, axis=0) + 1
    d = len(shapes)

    Y_curr = Y[idx[0]:idx[1]]
    Y_curr = Y_curr.reshape(shapes[0], -1, order='C')
    Y_curr, _ = skeleton(Y_curr, r=rank, eps=eps_skel)
    cores = [Y_curr[None, ...]]

    for mode in range(1, d):
        # The mode-th TT-core will have the shape r0 x n x r1
        r0 = cores[-1].shape[-1]
        r1 = rank if mode < d-1 else 1
        n = shapes[mode]

        I_curr = I[idx[mode]:idx[mode+1]]
        M = np.array([get_partial(cores[:mode], i) for i in I_curr[::idx_many[mode], :mode]])

        Y_curr = Y[idx[mode]:idx[mode+1]].reshape(-1, idx_many[mode], order='C')
        if Y_curr.shape[1] > r1:
            Y_curr, _ = skeleton(Y_curr, r=r1)
        r1 = Y_curr.shape[1]

        core = np.empty([r0, n, r1])
        step = Y_curr.shape[0] // n
        for i in range(n):
            A = M[i*step:(i+1)*step]
            b = Y_curr[i*step:(i+1)*step]
            core[:, i, :] = np.linalg.lstsq(A, b, rcond=-1)[0]
        cores.append(core)

    return cores



import numpy as np


from .maxvol import maxvol
from .maxvol import rect_maxvol
from .utils import ind2str


def cross(f, Y0, nswp=10, kr=2, rf=2, cache=None, info=None):
    d = len(Y0)
    Y = [Y0[i].copy() for i in range(d)]
    Il = [np.empty((1, 0), dtype=int)] + [None for i in range(d)]
    Ir = [None for i in range(d)] + [np.empty((1, 0), dtype=int)]
    Ig = [_reshape(np.arange(G.shape[1], dtype=int), (-1, 1)) for G in Y]

    R = np.ones((1, 1))
    for i in range(d-1):
        G = np.tensordot(R, Y[i], 1)
        Y[i], Il[i+1], R = cross_prep_l2r(G, Ig[i], Il[i])

    R = np.ones((1, 1))
    for i in range(d-1, 0, -1):
        G = np.tensordot(Y[i], R, 1)
        Y[i], Ir[i], R = cross_prep_r2l(G, Ig[i], Ir[i+1])

    def func(I):
        if cache is None:
            if info is not None:
                info['k_cache'] = 0
                info['k_evals'] = info.get('k_evals', 0) + len(I)
            return f(I)

        I_new = np.array([i for i in I if ind2str(i) not in cache])
        if len(I_new):
            Y_new = f(I_new)
            for k, i in enumerate(I_new):
                cache[ind2str(i)] = Y_new[k]

        if info is not None:
            info['k_cache'] = info.get('k_cache', 0) + (len(I) - len(I_new))
            info['k_evals'] = info.get('k_evals', 0) + len(I_new)

        Y = np.array([cache[ind2str(i)] for i in I])
        return Y

    for _ in range(nswp):
        R = np.ones((1, 1))
        for i in range(d):
            G = np.tensordot(R, Y[i], 1)
            y = func(cross_index_merge(Il[i], Ig[i], Ir[i+1]))
            Y[i], Il[i+1], R = cross_build_l2r(
                *G.shape, y, Ig[i], Il[i], kr, rf)
        Y[d-1] = np.tensordot(Y[d-1], R, 1)

        R = np.ones((1, 1))
        for i in range(d-1, -1, -1):
            G = np.tensordot(Y[i], R, 1)
            y = func(cross_index_merge(Il[i], Ig[i], Ir[i+1]))
            Y[i], Ir[i], R = cross_build_r2l(
                *G.shape, y, Ig[i], Ir[i+1], kr, rf)
        Y[0] = np.tensordot(R, Y[0], 1)

    return Y


def cross_build_l2r(r1, n, r2, y, Ig, I, kr, rf):
    G = _reshape(y, (-1, r2))
    Q, s, V = _svd(G)
    Ind, B = _maxvol_rect(Q, kr, rf)
    G = _reshape(B, (r1, n, -1))
    J = cross_index_stack_l2r(r1, Ig, I)[Ind, :]
    R = Q[Ind, :] @ np.diag(s) @ V
    return G, J, R


def cross_build_r2l(r1, n, r2, y, Ig, I, kr, rf):
    G = _reshape(y, (r1, -1)).T
    Q, s, V = _svd(G)
    Ind, B = _maxvol_rect(Q, kr, rf)
    G = _reshape(B.T, (-1, n, r2))
    J = cross_index_stack_r2l(r2, Ig, I)[Ind, :]
    R = (Q[Ind, :] @ np.diag(s) @ V).T
    return G, J, R


def cross_index_merge(i1, i2, i3):
    r1 = i1.shape[0] or 1
    r2 = i2.shape[0]
    r3 = i3.shape[0] or 1
    w1 = _kron(_ones(r3 * r2), i1)
    w2 = np.empty((w1.shape[0], 0))
    if i3.size and r2:
        w2 = _kron(i3, _ones(r1 * r2))
    w3 = _kron(_kron(_ones(r3), i2), _ones(r1))
    return np.hstack((w1, w3, w2))


def cross_index_stack_l2r(r, Ig, I):
    n = Ig.size
    J = _kron(Ig, _ones(r))
    if I.size:
        J = np.hstack((_kron(_ones(n), I), J))
    return _reshape(J, (n * r, -1))


def cross_index_stack_r2l(r, Ig, I):
    n = Ig.size
    J = _kron(_ones(r), Ig)
    if I.size:
        J = np.hstack((J, _kron(I, _ones(n))))
    return _reshape(J, (n * r, -1))


def cross_prep_l2r(G, Ig, I=None):
    r1, n, r2 = G.shape
    G = _reshape(G, (-1, G.shape[-1]))

    Q, R = _qr(G)
    Ind, B = _maxvol(Q)

    G = _reshape(B.T, (-1, n, r2))
    J = cross_index_stack_l2r(r1, Ig, I)[Ind, :]
    R = Q[Ind, :] @ R
    return G, J, R


def cross_prep_r2l(G, Ig, I=None):
    r1, n, r2 = G.shape
    G = _reshape(G, (G.shape[0], -1)).T

    Q, R = _qr(G)
    Ind, B = _maxvol(Q)

    G = _reshape(B.T, (-1, n, r2))
    J = cross_index_stack_r2l(r2, Ig, I)[Ind, :]
    R = (Q[Ind, :] @ R).T
    return G, J, R


def _kron(a, b):
    return np.kron(a, b)


def _maxvol(a):
    return maxvol(a)


def _maxvol_rect(a, kr=1, rf=1, tau=1.1):
    N, u = a.shape
    N_min = min(N, u + kr)
    N_max = min(N, u + kr + rf)

    if N <= u:
        I = np.arange(N, dtype=int)
        B = np.eye(N, dtype=a.dtype)
    else:
        I, B = rect_maxvol(a, tau, N_min, N_max)

    return I, B


def _ones(k, m=1):
    return np.ones((k, m), dtype=int)


def _qr(a):
    return np.linalg.qr(a)


def _reshape(a, shape):
    return np.reshape(a, shape, order='F')


def _svd(a, full_matrices=False):
    try:
        return np.linalg.svd(a, full_matrices=full_matrices)
    except:
        b = a + 1.E-14 * np.max(np.abs(a)) * np.random.randn(*a.shape)
        return np.linalg.svd(b, full_matrices=full_matrices)



__version__ = '0.7.1'


from .als import als
from .als import als2


from .anova import ANOVA


from .cross import cross


from .maxvol import maxvol
from .maxvol import rect_maxvol


from .tensor import add
from .tensor import erank
from .tensor import full
from .tensor import get
from .tensor import getter
from .tensor import mean
from .tensor import mul
from .tensor import norm
from .tensor import rand
from .tensor import show
from .tensor import sum
from .tensor import truncate


from .utils import confidence
from .utils import get_cdf
from .utils import ind2poi
from .utils import ind2str
from .utils import lhs
from .utils import str2ind
from .utils import tt_sample
from .utils import tt_svd
from .utils import tt_svd_incomplete



import numba as nb
import numpy as np
from copy import deepcopy


from .utils import orthogonalize
from .utils import reshape
from .utils import svd_truncated


def add(Y1, Y2):
    R1 = [1] + [G.shape[2] for G in Y1]
    R2 = [1] + [G.shape[2] for G in Y2]
    N = [G.shape[1] for G in Y1]
    Y = []
    for i, (G1, G2, n) in enumerate(zip(Y1, Y2, N)):
        if i == 0:
            G = np.concatenate([G1, G2], axis=2)
        elif i == len(N) - 1:
            G = np.concatenate([G1, G2], axis=0)
        else:
            R1_l, R1_r = R1[i:i+2]
            R2_l, R2_r = R2[i:i+2]
            Z1 = np.zeros([R1_l, n, R2_r])
            Z2 = np.zeros([R2_l, n, R1_r])
            L1 = np.concatenate([G1, Z1], axis=2)
            L2 = np.concatenate([Z2, G2], axis=2)
            G = np.concatenate([L1, L2], axis=0)
        Y.append(G)
    return Y


def erank(Y):
    """Compute effective rank of the TT-tensor."""
    d = len(Y)
    N = np.array([G.shape[1] for G in Y])
    R = np.array([1] + [G.shape[-1] for G in Y[:-1]] + [1])

    sz = np.dot(N * R[0:d], R[1:])
    b = R[0] * N[0] + N[d - 1] * R[d]
    a = np.sum(N[1:d - 1])
    return (np.sqrt(b * b + 4 * a * sz) - b) / (2 * a)


def full(Y):
    A = Y[0].copy()
    for i in range(1, len(Y)):
        A = np.tensordot(A, Y[i], 1)
    return A[0, ..., 0]


def get(Y, n):
    Q = Y[0][0, n[0], :]
    for i in range(1, len(Y)):
        Q = np.einsum('q,qp->p', Q, Y[i][:, n[i], :])
    return Q[0]


def getter(Y, compile=True):
    Y_nb = tuple([np.array(G, order='C') for G in Y])

    @nb.jit(nopython=True)
    def get(n):
        Q = Y_nb[0]
        y = [Q[0, n[0], r2] for r2 in range(Q.shape[2])]
        for i in range(1, len(Y_nb)):
            Q = Y_nb[i]
            R = np.zeros(Q.shape[2])
            for r1 in range(Q.shape[0]):
                for r2 in range(Q.shape[2]):
                    R[r2]+= y[r1] * Q[r1, n[i], r2]
            y = list(R)
        return y[0]

    if compile:
        y = get(np.zeros(len(Y), dtype=int))

    return get


def mean(Y, P=None, norm=True):
    """Compute mean value of the TT-tensor with the given probability."""
    R = np.ones((1, 1))
    for i in range(len(Y)):
        n = Y[i].shape[1]
        if P is not None:
            Q = P[i][:n]
        else:
            Q = np.ones(n) / n if norm else np.ones(n)
        R = R @ np.einsum('rmq,m->rq', Y[i], Q)
    return R[0, 0]


def mul(Y1, Y2):
    Y = []
    for G1, G2 in zip(Y1, Y2):
        G = G1[:, None, :, :, None] * G2[None, :, :, None, :]
        G = G.reshape([G1.shape[0]*G2.shape[0], -1, G1.shape[-1]*G2.shape[-1]])
        Y.append(G)
    return Y


def norm(Y):
    """Compute 2-norm of the given TT-tensor."""
    return np.sqrt(sum(mul(Y, Y)))


def rand(N, R, f=np.random.randn):
    N = np.asanyarray(N, dtype=np.int32)

    d = N.size

    if isinstance(R, (int, float)):
        R = [1] + [int(R)] * (d - 1) + [1]
    R = np.asanyarray(R, dtype=np.int32)

    ps = np.cumsum(np.concatenate(([1], N * R[0:d] * R[1:d+1])))
    ps = ps.astype(np.int32)
    core = f(ps[d] - 1)

    Y = []
    for i in range(d):
        G = core[ps[i]-1:ps[i+1]-1]
        Y.append(G.reshape((R[i], N[i], R[i+1]), order='F'))

    return Y


def show(Y):
    N = [G.shape[1] for G in Y]
    R = [G.shape[0] for G in Y] + [1]
    l = max(int(np.ceil(np.log10(np.max(R)+1))) + 1, 3)
    form_str = '{:^' + str(l) + '}'
    s0 = ' '*(l//2)
    s1 = s0 + ''.join([form_str.format(n) for n in N])
    s2 = s0 + ''.join([form_str.format('/ \\') for _ in N])
    s3 = ''.join([form_str.format(r) for r in R])
    print(f'{s1}\n{s2}\n{s3}\n')


def sum(Y):
    return mean(Y, norm=False)


def truncate(Y, e=1E-14, rmax=np.iinfo(np.int32).max, orth=True):
    d = len(Y)
    N = [G.shape[1] for G in Y]
    if orth:
        Z = orthogonalize(Y, d-1)
        delta = e / np.sqrt(d-1) * np.linalg.norm(Z[-1])
    else:
        Z = Y
        delta = e
    for k in range(d-1, 0, -1):
        M = reshape(Z[k], [Z[k].shape[0], -1])
        L, M = svd_truncated(M, delta, rmax)
        Z[k] = reshape(M, [-1, N[k], Z[k].shape[2]])
        Z[k-1] = np.einsum('ijk,kl', Z[k-1], L, optimize=True)
    return Z



# This is draft!


import jax
import jax.numpy as jnp
from jax.scipy.linalg import lu as jlu
from jax.scipy.linalg import solve_triangular as jsolve_triangular


@jax.jit
def maxvol(A, e=1.05, K=100):
    N, r = A.shape
    P, L, U = jlu(A)
    I = P.argmax(axis=0)
    Q = jsolve_triangular(U, A.T, trans=1, lower=False)
    B = jsolve_triangular(L[:r, :], Q, trans=1, unit_diagonal=True, lower=True)

    @jax.jit
    def step(args):
        I, B, i, j = args
        x = B[i, :]
        y = B[:, j]
        y = jax.ops.index_update(y, i, y[i] - 1)
        I = jax.ops.index_update(I, i, j)
        B-= jnp.outer(y / B[i, j], x)
        return I, B, i, j

    @jax.jit
    def cond(args):
        I, B, b, k = args
        return jnp.logical_and(k < K, jnp.abs(b) > e)

    @jax.jit
    def body(args):
        I, B, b, k = args
        i, j = jnp.divmod(jnp.abs(B).argmax(), N)
        b = B[i, j]
        I, B, i, j = jax.lax.cond(jnp.abs(b) > e, step,
            lambda args: args, operand=(I, B, i, j))
        return I, B, b, k+1

    I, B, b, k = jax.lax.while_loop(cond, body, (I, B, 2 * e, 0))
    return I[:r], B.T


def rect_maxvol(A, e, N_min, N_max, e0=1.05, K0=10): # DRAFT !!!
    N, r = A.shape
    I_tmp, B = maxvol(A, e0, K0)
    I = jnp.hstack([I_tmp, jnp.zeros(N_max-r, dtype=I_tmp.dtype)])
    S = jnp.ones(N, dtype=I_tmp.dtype)
    for s in I_tmp:
        S = jax.ops.index_update(S, s, 0)
    F = S * jnp.linalg.norm(B, axis=1)**2

    # @jax.jit
    def cond(args):
        I, B, F, S, f, k = args
        return jnp.logical_and(k < N_max, f > e*e)

    # @jax.jit
    def body(args):
        I, B, F, S, f, k = args
        i = jnp.argmax(F)
        # if k >= N_min and F[i] <= e*e: break
        I = jax.ops.index_update(I, k, i)
        S = jax.ops.index_update(S, i, 0)
        v = B.dot(B[i])
        l = 1. / (1 + v[i])
        B = jnp.hstack([B - l * jnp.outer(v, B[i]), l * v.reshape(-1, 1)])
        F = S * (F - l * v * v)
        return I, B, F, S, F[i], k+1

    I, B, F, S, f, k = jax.lax.while_loop(cond, body, (I, B, F, S, 2 * e*e, r))
    I = I[:B.shape[1]]
    #B[I] = jnp.eye(B.shape[1], dtype=B.dtype)
    return I, B



"""Approximation of the multivariate Rosenbrock function with noise."""
import numpy as np
from scipy.optimize import rosen
from time import perf_counter as tpc


import teneva


np.random.seed(1234567890)


# Parameters:
d         = 100         # Dimension of the function
A         = [-5.] * d   # Lower bound for spatial grid
B         = [+5.] * d   # Upper bound for spatial grid
N         = [100] * d   # Shape of the tensor (it may be non-uniform)
M_trn     = 100000      # Number of train points (for ANOVA and ALS)
M_tst     = 100000      # Number of test points
nswp      = 1           # Sweep number for cross
nswp_als  = 20          # Sweep number for ALS
eps       = 1.E-8       # Desired accuracy
kr        = 1           # Cross parameter (kickrank)
rf        = 1           # Cross parameter
order     = 1           # ANOVA order (1 or 2)
r0        = 3           # TT-rank for Cross initial guess
r         = 3           # TT-rank for ALS and ANOVA


# Target function:
def func(I):
    X = teneva.ind2poi(I, A, B, N)
    return rosen(X.T)


# Train data (for ALS and ANOVA):
I_trn = teneva.lhs(N, M_trn)
Y_trn = func(I_trn)


# Test data:
I_tst = np.vstack([np.random.choice(N[i], M_tst) for i in range(d)]).T
Y_tst = func(I_tst)


def als():
    Y = teneva.rand(N, r)

    Y = teneva.als(I_trn, Y_trn, Y, nswp_als)

    return Y, Y_trn.size, 0


def anova():
    anova = teneva.ANOVA(order, I_trn, Y_trn)
    Y = anova.cores(r)

    return Y, Y_trn.size, 0


def anova_als():
    anova = teneva.ANOVA(order, I_trn, Y_trn)
    Y = anova.cores(r)

    Y = teneva.als(I_trn, Y_trn, Y, nswp_als)

    return Y, Y_trn.size, 0


def anova_cross_cache():
    cache, info = {}, {}

    anova = teneva.ANOVA(order, I_trn, Y_trn)
    Y = anova.cores(r)

    Y = teneva.cross(func, Y, nswp, kr, rf, cache, info=info)
    Y = teneva.truncate(Y, eps)

    return Y, info['k_evals'], info['k_cache']


def anova_cross_cache_als():
    cache, info = {}, {}

    anova = teneva.ANOVA(order, I_trn, Y_trn)
    Y = anova.cores(r)

    Y = teneva.cross(func, Y, nswp, kr, rf, cache, info=info)
    Y = teneva.truncate(Y, eps)

    I_trn_new = np.array([teneva.str2ind(s) for s in cache.keys()], dtype=int)
    Y_trn_new = np.array([y for y in cache.values()])
    Y = teneva.als(I_trn_new, Y_trn_new, Y, nswp_als)

    return Y, Y_trn.size + info['k_evals'], info['k_cache']


def cross():
    info = {}

    Y = teneva.rand(N, r0)

    Y = teneva.cross(func, Y, nswp, kr, rf, info=info)
    Y = teneva.truncate(Y, eps)

    return Y, info['k_evals'], 0


def cross_cache():
    cache, info = {}, {}

    Y = teneva.rand(N, r0)

    Y = teneva.cross(func, Y, nswp, kr, rf, cache, info=info)
    Y = teneva.truncate(Y, eps)

    return Y, info['k_evals'], info['k_cache']


def cross_cache_als():
    cache, info = {}, {}

    Y = teneva.rand(N, r0)

    Y = teneva.cross(func, Y, nswp, kr, rf, cache, info=info)
    Y = teneva.truncate(Y, eps)

    I_trn_new = np.array([teneva.str2ind(s) for s in cache.keys()], dtype=int)
    Y_trn_new = np.array([y for y in cache.values()])
    Y = teneva.als(I_trn_new, Y_trn_new, Y, nswp_als)

    return Y, Y_trn.size + info['k_evals'], info['k_cache']


def proc(Y, k, k_cache, t_build, name):
    get = teneva.getter(Y)

    Z = np.array([get(i) for i in I_trn])
    e_trn = np.linalg.norm(Z - Y_trn) / np.linalg.norm(Y_trn)

    Z = np.array([get(i) for i in I_tst])
    e_tst = np.linalg.norm(Z - Y_tst) / np.linalg.norm(Y_tst)

    text = '\n'
    text += f'-------------- > {name}\n'
    text += f'Error on train : {e_trn:-10.2e}\n'
    text += f'Error on test  : {e_tst:-10.2e}\n'
    text += f'Tensor TT-rank : {teneva.erank(Y):-10.1f}\n'
    text += f'Build time     : {t_build:-10.2f}\n'
    text += f'Function evals : {k:-10d}'
    if k_cache:
        text += f' (+ cache {k_cache})'

    print(text)


def run():
    text = '\n'
    text += f'> Approximation of {d}-dim Rosenbrock function on uniform grid'
    print(text)

    t = tpc()
    Y, k, k_cache = anova()
    proc(Y, k, k_cache, tpc() - t, 'TT-ANOVA')

    t = tpc()
    Y, k, k_cache = als()
    proc(Y, k, k_cache, tpc() - t, 'TT-ALS')

    t = tpc()
    Y, k, k_cache = anova_als()
    proc(Y, k, k_cache, tpc() - t, 'TT-ANOVA + TT-ALS')

    t = tpc()
    Y, k, k_cache = cross()
    proc(Y, k, k_cache, tpc() - t, 'TT-Cross')

    t = tpc()
    Y, k, k_cache = cross_cache()
    proc(Y, k, k_cache, tpc() - t, 'TT-Cross-cache')

    t = tpc()
    Y, k, k_cache = anova_cross_cache()
    proc(Y, k, k_cache, tpc() - t, 'TT-ANOVA + TT-Cross-cache')

    t = tpc()
    Y, k, k_cache = cross_cache_als()
    proc(Y, k, k_cache, tpc() - t, 'TT-Cross-cache + TT-ALS')

    # t = tpc()
    # Y, k, k_cache = anova_cross_cache_als()
    # proc(Y, k, k_cache, tpc() - t, 'TT-ANOVA + TT-Cross-cache + TT-ALS')


if __name__ == '__main__':
    run()



import numpy as np


from .tensor import add
from .tensor import truncate


class ANOVA:
    def __init__(self, order, X=None, y=None):
        if not order in [1, 2]:
            raise ValueError('Invalid value for ANOVA order (should be 1 or 2')
        self.order = order

        if X is not None and y is not None:
            self.build(X, y)

    def build(self, X, y):
        self.dim = X.shape[1]
        self.domain = []
        self.shapes = np.zeros(self.dim, dtype=int)
        for i in range(self.dim):
            points = np.unique(X[:, i])
            self.domain.append(points)
            self.shapes[i] = len(points)

        self.build_0(X, y)
        if self.order >= 1:
            self.build_1(X, y)
        if self.order >= 2:
            self.build_2(X, y)

    def build_0(self, X, y):
        self.f0 = np.mean(y)

    def build_1(self, X, y):
        self.f1 = []
        self.f1_arr = []
        for i, dm in enumerate(self.domain):
            f1_curr = dict()
            f1_curr_arr = []
            for x in dm:
                value = np.mean(y[X[:, i] == x]) - self.f0
                f1_curr[x] = value
                f1_curr_arr.append(value)
            self.f1.append(f1_curr)
            self.f1_arr.append(np.array(f1_curr_arr))

    def build_2(self, X, y):
        self.f2 = []
        self.f2_arr = []
        for i1, dm1 in enumerate(self.domain[:-1]):
            for i2, dm2 in enumerate(self.domain[i1 + 1:], start=i1+1):
                f2_curr = dict()
                f2_curr_arr = []
                for x1 in dm1:
                    for x2 in dm2:
                        idx = (X[:, i1] == x1) & (X[:, i2] == x2)
                        if idx.sum() == 0:
                            #print(f"Warn: for (i1, i2)=({i1}, {i2}) there is no values")
                            value = 0.
                        else:
                            df = np.mean(y[idx]) - self.f0
                            value = df - self.f1[i1][x1] - self.f1[i2][x2]
                        f2_curr[(x1, x2)] = value
                        f2_curr_arr.append(value)
                self.f2.append(f2_curr)
                self.f2_arr.append(np.array(f2_curr_arr))

    def calc(self, X):
        res = self.calc_0()
        if self.order >= 1:
            res += self.calc_1(X)
        if self.order >= 2:
            res += self.calc_2(X)
        return res

    def calc_0(self):
        return self.f0

    def calc_1(self, X):
        res = 0.
        for i, x in enumerate(X):
            try:
                res += self.f1[i][x]
            except:
                print(f'Error: non val for f1[{i}][{x}]')
        return res

    def calc_2(self, X):
        res = 0.
        num = 0
        for i1, x1 in enumerate(X[:-1]):
            for x2 in X[i1 + 1:]:
                res += self.f2[num][(x1, x2)]
                num += 1
        return res

    def cores(self, rank=2, noise=1.E-10, only_near=False):
        if self.order < 1:
            raise ValueError('TT-cores may be constructed only if order >= 1')

        cores = self.cores_1(rank, noise)

        if self.order >= 2:
            cores2_many = self.cores_2(rank, only_near)
            cores = sum_many([cores] + cores2_many, rmax=rank)

        return cores

    def cores_1(self, rank=2, noise=1.E-10):
        cores = []

        core = noise * np.random.randn(1, self.shapes[0], rank)
        core[0, :, 0] = 1.
        core[0, :, 1] = self.f1_arr[0]
        cores.append(core)

        for i in range(1, self.dim-1):
            core = noise * np.random.randn(rank, self.shapes[i], rank)
            core[0, :, 0] = 1.
            core[1, :, 1] = 1.
            core[0, :, 1] = self.f1_arr[i]
            cores.append(core)

        core = noise * np.random.randn(rank, self.shapes[self.dim-1], 1)
        core[0, :, 0] = self.f1_arr[self.dim-1] + self.f0
        core[1, :, 0] = 1.
        cores.append(core)

        return cores

    def cores_2(self, rank=2, only_near=False):
        mats = []
        num = 0
        for i1 in range(self.dim-1):
            for i2 in range(i1+1, self.dim):
                shape = (self.shapes[i1], self.shapes[i2])
                mat = self.f2_arr[num].reshape(shape, order='C')
                mats.append(mat)
                num += 1

        cores = []
        num = 0
        for i1 in range(self.dim-1):
            for i2 in [i1+1] if only_near else range(i1+1, self.dim):
                cores.append(second_order_2_TT(mats[num], i1, i2, self.shapes))
                num += 1

        return cores

    def __call__(self, X):
        X = np.asanyarray(X)
        if len(X.shape) == 1:
            return self.calc(X)
        if len(X.shape) == 2:
            return np.array([self.calc(x) for x in X])
        return None

    def __getitem__(self, X):
        return self(X)


def skeleton(a, eps=1.E-10, r=int(1e10), hermitian=False):
    u, s, v = np.linalg.svd(a, full_matrices=False,
        compute_uv=True, hermitian=hermitian)
    r = min(r, sum(s/s[0] > eps))
    un = u[:, :r]
    sn = np.diag(np.sqrt(s[:r]))
    vn = v[:r]
    return un @ sn, sn @ vn


def core_one(n, r):
    core = np.zeros([r, n, r])
    for x in range(n):
        core[range(r), x, range(r)] = 1.
    return core


def second_order_2_TT(A, i, j, shapes):
    if i > j: # Так не должно быть
        j, i = i, j
        A = A.T

    U, V = skeleton(A)
    r = U.shape[1]

    core1 = U.reshape(1, U.shape[0], r)
    core2 = V.reshape(r, V.shape[1], 1)

    cores = []
    for num, n in enumerate(shapes):
        if num < i or num > j:
            cores.append(np.ones([1, n, 1]))
        if num == i:
            cores.append(core1)
        if num == j:
            cores.append(core2)
        if i < num < j:
            cores.append(core_one(n, r))

    return cores


def sum_many(tensors, e=1.E-10, rmax=None, freq_trunc=15):
    cores = tensors[0]
    for i, t in enumerate(tensors[1:]):
        cores = add(cores, t)
        if (i+1) % freq_trunc == 0:
            cores = truncate(cores, e=e)
    return truncate(cores, e=e, rmax=rmax)



import numpy as np
from scipy.linalg import lu
from scipy.linalg import solve_triangular


def maxvol(A, e=1.05, k=100):
    n, r = A.shape
    assert e >= 1 and n > r and k > 0
    P, L, U = lu(A, check_finite=False)
    I = P[:, :r].argmax(axis=0)
    Q = solve_triangular(U, A.T, trans=1, check_finite=False)
    B = solve_triangular(L[:r, :], Q, trans=1, check_finite=False,
                         unit_diagonal=True, lower=True).T
    for _ in range(k):
        i, j = np.divmod(np.abs(B).argmax(), r)
        if np.abs(B[i, j]) <= e: break
        I[j] = i
        bj = B[:, j]
        bi = B[i, :].copy()
        bi[j] -= 1.
        B -= np.outer(bj, bi / B[i, j])
    return I, B


def rect_maxvol(A, e, N_min, N_max, e0=1.05, K0=10):
    N, r = A.shape
    assert e >= 1 and N > r and N_min >= r and N_min <= N_max and N_max <= N
    I_tmp, B = maxvol(A, e0, K0)
    I = np.hstack([I_tmp, np.zeros(N_max-r, dtype=I_tmp.dtype)])
    S = np.ones(N, dtype=np.int32)
    S[I_tmp] = 0
    F = S * np.linalg.norm(B, axis=1)**2
    for k in range(r, N_max):
        i = np.argmax(F)
        if k >= N_min and F[i] <= e*e: break
        I[k] = i
        S[i] = 0
        v = B.dot(B[i])
        l = 1. / (1 + v[i])
        B = np.hstack([B - l * np.outer(v, B[i]), l * v.reshape(-1, 1)])
        F = S * (F - l * v * v)
    I = I[:B.shape[1]]
    B[I] = np.eye(B.shape[1], dtype=B.dtype)
    return I, B
